#include "Enemy.h"

void clearEnemies()
{
    roomEnemyCount = 0;

    memset(roomEnemies, 0, sizeof(Enemy) * MAX_ROOM_ENEMY_COUNT);
}

void addEnemy(u8 x, u8 y, u8 type)
{
    u8 nextEnemy = nextEnemyIndex();

    if (nextEnemy < MAX_ROOM_ENEMY_COUNT)
    {
        initEnemy(&roomEnemies[nextEnemy], x, y, type);

        // The count needs to be incremented after adding each new enemy.
        if (roomEnemyCount < MAX_ROOM_ENEMY_COUNT)
        {
            ++roomEnemyCount;
        }
    }
}

void initEnemy(Enemy* enemy, u8 x, u8 y, u8 type)
{
    enemy->object.position.x = (x % SCREEN_TILES_X) * TILE_WIDTH_BYTES;
    enemy->object.position.y = (y % SCREEN_TILES_Y) * TILE_HEIGHT;
    enemy->object.previousPosition.x = enemy->object.position.x;
    enemy->object.previousPosition.y = enemy->object.position.y;
    enemy->type = type;
    enemy->object.width = TILE_WIDTH_BYTES;
    enemy->object.height = TILE_HEIGHT;
    enemy->id = nextEnemyIndex();

    // Completely arbitrary velocity.
    enemy->object.velocity.x = 1;

    // Using health to determine if an enemy is alive or not.
    enemy->health = 1;
}

u8 nextEnemyIndex()
{

    u8 i = 0; 

    for (i = 0; i < MAX_ROOM_ENEMY_COUNT; ++i)
    {
        if (roomEnemies[i].health == 0)
        {
            return i;
        }
    }

    // No space for any more enemies.
    return 255;
}

void drawEnemies()
{
    u8 i = 0;
    Enemy* enemy = roomEnemies;

    for (i = 0; i < roomEnemyCount; ++i)
    {
        // Using health to determine if enemy has been destroyed.
        if (enemy->redraw && enemy->health)
        {
            drawActor((Actor*)enemy, allegroEnemySprites[enemy->type]);
            enemy->redraw = FALSE;
        }

        // Move to the next enemy.
        ++enemy;
    }
}

void redrawAllEnemies()
{
    u8 i = 0;
    Enemy* enemy = roomEnemies;

    for (i = 0; i < roomEnemyCount; ++i)
    {
        // Only redraw enemies that are valid.
        if (enemy->health)
        {
            enemy->redraw = TRUE;
        }

        // Move to the next enemy.
        ++enemy;
    }
}

void logicEnemies()
{
    u8 i = 0;
    Enemy* enemy = roomEnemies;

    for (i = 0; i < roomEnemyCount; ++i)
    {
        // Only run logic for enemies that have health (alive and on screen).
        if (enemy->health)
        {
            // Check if the enemy is moving.
            if (enemy->object.velocity.x || enemy->object.velocity.y)
            {
                // Using the logic flag to reduce the frequency of running the enemy logic to improve performance.
                // This results in the enemy being drawn less, unfortunately the enemy looks less smooth doing this as well.
                enemy->logicFlag++;

                if (enemy->logicFlag > 2)
                {
                    updateEnemyPosition(enemy);

                    enemy->logicFlag = 0;

                    // Code for adjusting enemy based on player position previously went here.
                }
            }
        }

        // Move to next enemy.
        ++enemy;
    }
}

void updateEnemyPosition(Enemy* enemy)
{
    // The tiles that the enemy currently occupies.
    u8 occupiedXTileLeft = (enemy->object.position.x) >> 2;
    u8 occupiedXTileRight = (enemy->object.position.x + TILE_WIDTH_BYTES - 1);
    u8 occupiedYTileBottom = (enemy->object.position.y + TILE_HEIGHT - 1);
    u8 occupiedYTileTop = (enemy->object.position.y) >> 4;
    occupiedXTileRight >>= 2;
    occupiedYTileBottom >>= 4;

    // Save the previous position of the enemy.
    enemy->object.previousPosition.x = enemy->object.position.x;
    enemy->object.previousPosition.y = enemy->object.position.y;

    // Process horizontal enemy movement.
    if (enemy->object.velocity.x)
    {
        u8 futureXTileLeft = (enemy->object.position.x + enemy->object.velocity.x) >> 2;
        u8 futureXTileRight = (enemy->object.position.x + TILE_WIDTH_BYTES - 1 + enemy->object.velocity.x);
        futureXTileRight >>= 2;

        if (enemy->object.velocity.x > 0 && enemy->object.position.x + enemy->object.velocity.x <= SCREEN_X - TILE_WIDTH_BYTES)
        {
            // Enemy going right.
            if (processCellForEnemy(futureXTileRight, occupiedYTileTop) == CELL_EMPTY &&
                processCellForEnemy(futureXTileRight, occupiedYTileBottom) == CELL_EMPTY)
            {
                enemy->object.position.x += enemy->object.velocity.x;
            }
            else
            {
                // Currently causes an enemy get stuck on test map 3.
                // Enemy blocked so change direction to go up or down.
                if (enemy->id % 2)
                {
                    enemy->object.velocity.y = -3;
                }
                else
                {
                    enemy->object.velocity.y = 3;
                }
            }
        }
        else if (enemy->object.velocity.x > 0 && enemy->object.position.x + enemy->object.velocity.x >= SCREEN_X - TILE_WIDTH_BYTES)
        {
            // Enemy reached bottom of screen so destroy the enemy.
            enemy->health = 0;
        }
        else if (enemy->object.velocity.x < 0 && enemy->object.position.x + enemy->object.velocity.x >= 0)
        {
            // Enemy going left.
            if (processCellForEnemy(futureXTileLeft, occupiedYTileTop) == CELL_EMPTY &&
                processCellForEnemy(futureXTileLeft, occupiedYTileBottom) == CELL_EMPTY)
            {
                enemy->object.position.x += enemy->object.velocity.x;
            }
            else
            {
                // Enemy blocked so change direction.
                enemy->object.velocity.x *= -1;
            }
        }
    }

    // Process vertical movement after horizontal movement.
    if (enemy->object.velocity.y)
    {
        u8 futureYTileBottom = (enemy->object.position.y + TILE_HEIGHT - 1 + enemy->object.velocity.y);
        u8 futureYTileTop = (enemy->object.position.y + enemy->object.velocity.y) >> 4;
        futureYTileBottom >>= 4;

        // Recalculate occupied x position following the change of position after processing x velocity.
        // This prevents the enemy from clipping through solid tiles.
        occupiedXTileLeft = (enemy->object.position.x) >> 2;
        occupiedXTileRight = (enemy->object.position.x + TILE_WIDTH_BYTES - 1);
        occupiedXTileRight >>= 2;

        if (enemy->object.velocity.y > 0 && enemy->object.position.y + enemy->object.velocity.y <= SCREEN_Y_NO_HUD - TILE_HEIGHT)
        {
            // Enemy going down.
            if (processCellForEnemy(occupiedXTileLeft, futureYTileBottom) == CELL_EMPTY &&
                processCellForEnemy(occupiedXTileRight, futureYTileBottom) == CELL_EMPTY)
            {
                enemy->object.position.y += enemy->object.velocity.y;
            }
            else
            {
                // Enemy blocked so change direction.
                enemy->object.velocity.y *= -1;
            }
        }
        else if (enemy->object.velocity.y > 0 && enemy->object.position.y + enemy->object.velocity.y >= SCREEN_Y_NO_HUD - TILE_HEIGHT)
        {
            // Reached the bottom end of the screen so destroy the enemy.
            enemy->health = 0;
        }
        else if (enemy->object.velocity.y < 0 && enemy->object.position.y + enemy->object.velocity.y >= 0)
        {
            // Enemy going up.
            if (processCellForEnemy(occupiedXTileLeft, futureYTileTop) == CELL_EMPTY &&
                processCellForEnemy(occupiedXTileRight, futureYTileTop) == CELL_EMPTY)
            {
                enemy->object.position.y += enemy->object.velocity.y;
            }
            else
            {
                // Enemy blocked so change direction.
                enemy->object.velocity.y *= -1;
            }
        }
        else if (enemy->object.velocity.y < 0 && enemy->object.position.y + enemy->object.velocity.y <= 0)
        {
            // Reached the top end of the screen so destroy the enemy.
            enemy->health = 0;
        }
    }

    // If the enemy has moved it needs to be redrawn.
    if (enemy->object.previousPosition.x != enemy->object.position.x ||
        enemy->object.previousPosition.y != enemy->object.position.y)
    {
        enemy->redraw = TRUE;
    }
}

u8 processCellForEnemy(u8 x, u8 y)
{
    // I don't think we will have additional rooms.
    u8 currentRoomX = 0;
    u8 currentRoomY = 0;
    u8 cell = mapCell(x, y, currentRoomX, currentRoomY);

    // Empty cells treated as non-solid.
    if (cell == 129 || cell == CELL_EMPTY)
    {
        return CELL_EMPTY;
    }

    // Everything else is treated as solid.
    return 1;
}

u16 fishScore(u8 type)
{
    switch (type)
    {
    case 0:
        // Slug gives you almost nothing.
        return 1;

    case 1:
        // Big grey fish.
        return 100;

    case 2:
        // Big stripy fish.
        return 150;

    case 3:
    case 4:
        // Little fish.
        return 2;

    case 5:
        // Medium fish.
        return 20;

    case 6:
        // Octopus.
        return 250;

    case 7:
        // Crab.
        return 400;
    }

    // Everything else gives you nothing.
    return 0;
}