#include "Game.h"

// The allegro timer used for the remaining time in the level.
ALLEGRO_TIMER* levelTimer_ = NULL;

// The allegro timer used for determining when fish should be spawned in the level.
ALLEGRO_TIMER* fishTimer_ = NULL;

void runGameSession(u8 fpsCap)
{
    // Clear everything off the screen.
    al_clear_to_color(al_map_rgb(0, 0, 128));
    al_flip_display();

    // Reset game state values for a new game.
    resetGameSessionState();

    // Set the game to running for the game loop.
    gameRunning = TRUE;

    // Loop while level state is ready.
    while (!levelState)
    {
        // Initialise the map from the level array.
        initMap();

        // Initialise the player values, with map position.
        initPlayer(6, 6);

        // Clear any enemies before starting the map.
        clearEnemies();

        // Add some test enemies
        addEnemy(4, 4, 4);
        addEnemy(9, 7, 4);

        // Initially drawing the hud.
        hudRedraw = TRUE;

        // Run the actual game including the game loop.
        runGame(fpsCap);

        if (currentMap < LEVEL_COUNT)
        {
            // Update current map to prepare to move to the next level.
            ++currentMap;

            // Check if all the levels have been completed.
            if (currentMap == LEVEL_COUNT)
            {
                // Reset current map back to 0.
                currentMap = 0;

                // If currentMap is the last level then game complete
                break;
            }
            else
            {
                // Move to the next level.
                gameRunning = TRUE;
            }
        }
    }
}

void runGame(u8 fpsCap)
{
    u8 logicFlag = 0;
    const char logicSpeed = 60;
    ALLEGRO_TIMER* timer = al_create_timer(1.0 / logicSpeed);
    ALLEGRO_EVENT allegroEvent;

    ALLEGRO_EVENT_QUEUE* gameEventQueue = al_create_event_queue();
    al_register_event_source(gameEventQueue, al_get_keyboard_event_source());
    al_register_event_source(gameEventQueue, al_get_timer_event_source(timer));
    al_register_event_source(gameEventQueue, al_get_display_event_source(al_get_current_display()));

    // Setup timers required for the game.
    initialiseAllAllegroTimers();

    // Make sure any down keys are reset before starting the next game.
    gameControls.controlFlags = 0;

    // Start the timer that will ensure the logic is at a consistent speed.
    al_start_timer(timer);

    // Start the level timer that will count down the level time.
    al_start_timer(levelTimer_);

    // Start the fish timer to start spawning additional fish.
    al_start_timer(fishTimer_);

    while (gameRunning)
    {
        al_wait_for_event(gameEventQueue, &allegroEvent);

        // Read input from the player.
        readInput(gameEventQueue, &allegroEvent);

        // Only run logic every second time.
        if (logicFlag++)
        {
            // Run the game logic which will change what is displayed.
            gameLogic();

            // Reset flag
            logicFlag = 0;
        }

        // Check the timers used in game if they have elapsed.
        checkTimers();

        if (allegroEvent.type == ALLEGRO_EVENT_TIMER)
        {
            // Redraw every loop.
            // TODO consider backbuffer for allegro to reduce redraw.
            mapRedraw = TRUE;
            player.redraw = TRUE;
            redrawAllEnemies();
        }

        while ((fpsCap == FALSE || mapRedraw) && al_is_event_queue_empty(gameEventQueue))
        {
            // Draws the game on the screen.
            drawGame();

            al_flip_display();
        }
    }
}

void drawGame()
{
    // Draws the map grid on the screen.
    drawMap();

    // Draw the enemies on the screen.
    drawEnemies();

    // Draw the player on the screen at their current position (should be drawn on top of any fish).
    drawPlayer();

    // Draw the hud at the bottom of the screen.
    drawHud();
}

void readInput(ALLEGRO_EVENT_QUEUE* eventQueue, ALLEGRO_EVENT* allegroEvent)
{
    // Reads the game controls from the player.
    readGameControls(allegroEvent);

    // Check that the exit key has been pressed.
    if (gameControls.controlFlags & CONTROL_EXIT_FLAG)
    {
        // Stop the timer while the exit dialog is shown.
        al_stop_timer(levelTimer_);

        // If the player chooses to press yes exit the game.
        if (exitDialog(eventQueue, allegroEvent))
        {
            gameRunning = FALSE;
            levelState = LEVEL_EXIT;
        }
        else
        {
            al_clear_to_color(al_map_rgb(0, 0, 128));
            al_flip_display();

            // Force a map redraw to cleanup after the dialog is closed.
            mapRedraw = TRUE;

            // Force a redraw of the hud.
            hudRedraw = TRUE;
        }

        // Resume the timer once the exit dialog has been closed.
        al_resume_timer(levelTimer_);
    }
}

void gameLogic()
{
    // Move the player (update velocity).
    movePlayer();

    // Run the logic for the enemies (update position and velocity).
    logicEnemies();

    // Run the logic for the player (update position).
    logicPlayer();

    // Spawns new fish in the level to eat based on the fish timer.
    spawnFish();
}

void initialiseAllAllegroTimers()
{
    levelTimer_ = al_create_timer(1);
    fishTimer_ = al_create_timer(0.7);
}

void destroyAllAllegroTimers()
{
    al_destroy_timer(levelTimer_);
    al_destroy_timer(fishTimer_);
}

void checkTimers()
{
    int currentSecondsElapsed = al_get_timer_count(levelTimer_);

    if (currentSecondsElapsed > levelTimerElapsedSeconds)
    {
        // Updates the game timer on the hud.
        timeRedraw = TRUE;

        // Update seconds elapsed with the new time.
        levelTimerElapsedSeconds = currentSecondsElapsed;

        if (currentSecondsElapsed == MAX_LEVEL_TIME)
        {
            // End the game once the timer runs out.
            levelState = LEVEL_TIME_OUT;
            gameRunning = FALSE;
        }
    }

    int currentFishTimerCount = al_get_timer_count(fishTimer_);

    if (currentFishTimerCount > fishTimerCount)
    {
        fishTimerElapsed = TRUE;

        fishTimerCount = currentFishTimerCount;
    }
}

void spawnFish()
{
    if (fishTimerElapsed)
    {
        u8 x = 0;
        u8 y = 0;
        nextFishCoordinates(&x, &y);

        u8 nextFish = nextFishType();
        addEnemy(x, y, nextFish);

        fishTimerElapsed = FALSE;
    }
}

void nextFishCoordinates(u8* x, u8* y)
{
    u8 nextX = 0;
    u8 nextY = 0;

    srand(fishTimerCount);

    // Check that the randomly generated coordinates are not blocked.
    do
    {
        nextX = rand() % SCREEN_TILES_X;
        nextY = rand() % SCREEN_TILES_Y;
    } while (mapCell(nextX, nextY, 0, 0) != 129);

    *x = nextX;
    *y = nextY;
}

u8 nextFishType()
{
    srand(levelTimerElapsedSeconds);

    if (fishTimerCount % 3 == 0)
    {
        // 0 -> 7
        return rand() % 8;
    }
    else if (fishTimerCount % 3 == 1)
    {
        // 0 -> 3
        return rand() % 4;
    }
    else
    {
        // 3 -> 4
        // Default to basic fish.
        return 3 + (rand() % 2);
    }
}