#include "Menu.h"

// The colour of the text on the menus.
ALLEGRO_COLOR textColour_;

// Menu item number controls.
#define START_GAME_KEY_ ALLEGRO_KEY_1
#define REBIND_CONTROLS_KEY_ ALLEGRO_KEY_2
#define HIGH_SCORES_KEY_ ALLEGRO_KEY_3
#define EXIT_KEY_ ALLEGRO_KEY_4

// High score variables.
// The number of high score entries.
#define SCORE_COUNT_ 5

// The maximum length that a high score name can be.
#define MAX_NAME_LENGTH_ 10

// The high score values in the high score list.
unsigned int highScores_[SCORE_COUNT_];

// The names for the high score values in the high score list.
unsigned char highScoreNames_[SCORE_COUNT_][11];

void runMenu(u8 fpsCap)
{
    int menuRunning = TRUE;
    ALLEGRO_EVENT_QUEUE* menuEventQueue = NULL;
    ALLEGRO_EVENT allegroEvent;
    textColour_ = al_map_rgb(255, 255, 255);

    // Set the default high score entries.
    initHighScores();

    // Set the default keys for the player.
    setDefaultKeys();

    // Draw the menu on screen.
    drawMenu();

    setupMenuEventQueue(&menuEventQueue);

    while (menuRunning)
    {
        // Updates the content of the screen buffers.
        al_flip_display();

        al_wait_for_event(menuEventQueue, &allegroEvent);

        // Get the user chosen key.
        unsigned short key = getKey(&menuEventQueue);

        if (key == START_GAME_KEY_)
        {
            // Runs the full game session.
            runFullGame(&menuEventQueue, fpsCap);

            // Redraw the menu after finishing the game session.
            drawMenu();
        }
        else if (key == REBIND_CONTROLS_KEY_)
        {
            // Display rebind controls menu which asks the user for their chosen controls.
            rebindControls(&menuEventQueue);

            // Redraw the menu after leaving rebind controls menu.
            drawMenu();
        }
        else if (key == HIGH_SCORES_KEY_)
        {
            // Displays the high scores on the screen.
            displayHighScores(&menuEventQueue);

            // Redraw the menu after leaving the high scores.
            drawMenu();
        }
        else if (key == EXIT_KEY_)
        {
            menuRunning = FALSE;
        }
    }

    // Cleanup
    al_destroy_event_queue(menuEventQueue);
}

void drawMenu()
{
    // Clear the screen before drawing the menu.
    al_clear_to_color(al_map_rgb(0, 0, 128));

    // Game title.
    drawString(104 * scaleFactor, 30 * scaleFactor, "TINS 2023", textColour_);

    // Menu item choices.
    drawString(80 * scaleFactor, 60 * scaleFactor, "1. START GAME", textColour_);
    drawString(80 * scaleFactor, 75 * scaleFactor, "2. REBIND CONTROLS", textColour_);
    drawString(80 * scaleFactor, 90 * scaleFactor, "3. HIGH SCORES", textColour_);
    drawString(80 * scaleFactor, 105 * scaleFactor, "4. EXIT", textColour_);

    // Draw the game version number on the bottom right corner of the screen.
    drawString(220 * scaleFactor, 160 * scaleFactor, STRFILEVER, textColour_);
}

void setupMenuEventQueue(ALLEGRO_EVENT_QUEUE** eventQueue)
{
    al_install_keyboard();

    const char logicSpeed = 60;
    ALLEGRO_TIMER* timer = al_create_timer(1.0 / logicSpeed);

    // Setup the event queue and the event sources.
    *eventQueue = al_create_event_queue();
    al_register_event_source(*eventQueue, al_get_keyboard_event_source());
    al_register_event_source(*eventQueue, al_get_timer_event_source(timer));
    al_register_event_source(*eventQueue, al_get_display_event_source(al_get_current_display()));

    // Start the timer that will ensure the logic is at a consistent speed.
    al_start_timer(timer);
}

unsigned short int getKey(ALLEGRO_EVENT_QUEUE** menuEventQueue)
{
    ALLEGRO_EVENT allegroEvent;
    unsigned short int keycode = ALLEGRO_KEY_MAX;

    // Flush the event queue to prevent previous keyboard events being caught.
    al_flush_event_queue(*menuEventQueue);

    while (keycode == ALLEGRO_KEY_MAX)
    {
        al_wait_for_event(*menuEventQueue, &allegroEvent);

        if (allegroEvent.type == ALLEGRO_EVENT_KEY_DOWN)
        {
            keycode = allegroEvent.keyboard.keycode;
        }
        else if (allegroEvent.type == ALLEGRO_EVENT_KEY_UP)
        {
            if (keycode != ALLEGRO_KEY_MAX)
            {
                return keycode;
            }
        }
    }

    return keycode;
}

void displayHighScores(ALLEGRO_EVENT_QUEUE** eventQueue)
{
    unsigned char i = 0;
    unsigned char scoreStart = 25;

    al_clear_to_color(al_map_rgb(0, 0, 128));

    // High scores title.
    drawString(104 * scaleFactor, 30 * scaleFactor, "HIGH SCORES", textColour_);

    // Each of the entires in the high score list.
    for (i = 0; i < SCORE_COUNT_; ++i)
    {
        // Draw the name of the player from the names array.
        unsigned char y = 65 + i * 15;
        drawString(160 * scaleFactor, y * scaleFactor, highScoreNames_[i], textColour_);

        // Draw the score with the zeros (there is an extra 0 so that scores can be bigger than the are actually stored).
        drawScore(highScores_[i], 104 * scaleFactor, y * scaleFactor, textColour_);
    }

    al_flip_display();
    getKey(eventQueue);
}

void initHighScores()
{
    unsigned char i = 0;

    for (i = 0; i < SCORE_COUNT_; ++i)
    {
        highScores_[i] = 10;

        memcpy(highScoreNames_[i], "HERON", 6);
    }
}

void rebindControls(ALLEGRO_EVENT_QUEUE** eventQueue)
{
    // Clear the screen so that we can show control text.
    al_clear_to_color(al_map_rgb(0, 0, 128));
    al_flip_display();

    // Enter controls title.
    drawString(104 * scaleFactor, 30 * scaleFactor, "ENTER CONTROLS", textColour_);

    // Draw the text indicating which control is being set.
    // Make sure no key is currently down.
    // Set the key for the next control (this is done for each of control below).
    drawString(130 * scaleFactor, 50 * scaleFactor, "UP", textColour_);
    al_flip_display();
    gameControls.up = getKey(eventQueue);

    drawString(130 * scaleFactor, 65 * scaleFactor, "DOWN", textColour_);
    al_flip_display();
    gameControls.down = getKey(eventQueue);

    drawString(130 * scaleFactor, 80 * scaleFactor, "LEFT", textColour_);
    al_flip_display();
    gameControls.left = getKey(eventQueue);

    drawString(130 * scaleFactor, 95 * scaleFactor, "RIGHT", textColour_);
    al_flip_display();
    gameControls.right = getKey(eventQueue);

    drawString(130 * scaleFactor, 110 * scaleFactor, "ACTION", textColour_);
    al_flip_display();
    gameControls.action = getKey(eventQueue);

    drawString(130 * scaleFactor, 125 * scaleFactor, "EXIT", textColour_);
    al_flip_display();
    gameControls.exit = getKey(eventQueue);
}

void runFullGame(ALLEGRO_EVENT_QUEUE** eventQueue, u8 fpsCap)
{
    // Runs the full game session.
    runGameSession(fpsCap);

    if (player.fishEaten)
    {
        // When the player has eaten fish there is a chart showing what fish they have eaten.
        displayFishChart(eventQueue);
    }

    // Update high scores.
    updateHighScores(eventQueue);

    // Display high scores after entering high score.
    displayHighScores(eventQueue);
}

void updateHighScores(ALLEGRO_EVENT_QUEUE** eventQueue)
{
    // Check that the player got a higher score than the bottom score.
    if (levelScore > highScores_[4])
    {
        // Let the player enter their name.
        static u8 scoreName[11];
        u8 i = 0;

        // Set scoreName to the null terminator to end the string.
        memset(scoreName, '\0', 11);

        // Clear the screen
        al_clear_to_color(al_map_rgb(0, 0, 128));

        // Draw the score with padding of zeroes.
        drawString(65 * scaleFactor, 40 * scaleFactor, "NEW HIGH SCORE", textColour_);
        drawScore(levelScore, 185 * scaleFactor, 40 * scaleFactor, textColour_);
        drawString(80 * scaleFactor, 60 * scaleFactor, "ENTER YOUR NAME", textColour_);

        // Enter the name of the user for the high scores.
        al_flip_display();
        enterWord(scoreName, MAX_NAME_LENGTH_, eventQueue);


        // Update score values.
        highScores_[4] = levelScore;
        memcpy(highScoreNames_[4], scoreName, 11);

        for (i = 4; i > 0; --i)
        {
            if (highScores_[i] > highScores_[i - 1])
            {
                highScores_[i] = highScores_[i - 1];
                memcpy(highScoreNames_[i], highScoreNames_[i - 1], 11);

                highScores_[i - 1] = levelScore;
                memcpy(highScoreNames_[i - 1], scoreName, 11);
            }
        }
    }
}

void enterWord(u8* word, u8 maxLength, ALLEGRO_EVENT_QUEUE** eventQueue)
{
    // Default the chosen key to whatever.
    unsigned short int chosenKey = ALLEGRO_KEY_0;
    u8 letterCount = 0;
    u8 letter = 0;
    u8 wordChange = FALSE;

    do
    {
        // Get the next character from the user.
        chosenKey = getKey(eventQueue);

        // Del to move back to the previous key.
        if (chosenKey == ALLEGRO_KEY_DELETE || chosenKey == ALLEGRO_KEY_BACKSPACE)
        {
            if (letterCount)
            {
                // Remove the letter from the word array.
                --letterCount;
                word[letterCount] = '\0';

                // Indicate a word change to draw it on screen.
                wordChange = TRUE;
            }
        }

        // Make sure the key is in the correct range and the word is not too long.
        else if (letterCount < maxLength)
        {
            // Convert the chosen key (cpct_keyID to a character to an ASCII character).
            switch (chosenKey)
            {
            case ALLEGRO_KEY_A:
                letter = 'A';
                break;

            case ALLEGRO_KEY_B:
                letter = 'B';
                break;

            case ALLEGRO_KEY_C:
                letter = 'C';
                break;

            case ALLEGRO_KEY_D:
                letter = 'D';
                break;

            case ALLEGRO_KEY_E:
                letter = 'E';
                break;

            case ALLEGRO_KEY_F:
                letter = 'F';
                break;

            case ALLEGRO_KEY_G:
                letter = 'G';
                break;

            case ALLEGRO_KEY_H:
                letter = 'H';
                break;

            case ALLEGRO_KEY_I:
                letter = 'I';
                break;

            case ALLEGRO_KEY_J:
                letter = 'J';
                break;

            case ALLEGRO_KEY_K:
                letter = 'K';
                break;

            case ALLEGRO_KEY_L:
                letter = 'L';
                break;

            case ALLEGRO_KEY_M:
                letter = 'M';
                break;

            case ALLEGRO_KEY_N:
                letter = 'N';
                break;

            case ALLEGRO_KEY_O:
                letter = 'O';
                break;

            case ALLEGRO_KEY_P:
                letter = 'P';
                break;

            case ALLEGRO_KEY_Q:
                letter = 'Q';
                break;

            case ALLEGRO_KEY_R:
                letter = 'R';
                break;

            case ALLEGRO_KEY_S:
                letter = 'S';
                break;

            case ALLEGRO_KEY_T:
                letter = 'T';
                break;

            case ALLEGRO_KEY_U:
                letter = 'U';
                break;

            case ALLEGRO_KEY_V:
                letter = 'V';
                break;

            case ALLEGRO_KEY_W:
                letter = 'W';
                break;

            case ALLEGRO_KEY_X:
                letter = 'X';
                break;

            case ALLEGRO_KEY_Y:
                letter = 'Y';
                break;

            case ALLEGRO_KEY_Z:
                letter = 'Z';
                break;

            case ALLEGRO_KEY_SPACE:
                letter = ' ';
                break;
            }

            if (letter)
            {
                // Add the chosen letter to the string that makes up the word that the user is entering.
                word[letterCount] = letter;

                // Indicate a word change to draw it on screen.
                wordChange = TRUE;

                // Indicate a new letter has been added to the word array.
                ++letterCount;

                // Make sure the chosen character is reset so the previous character
                // does not get drawn if the user chooses an invalid character.
                letter = 0;
            }
        }

        // Draw the whole word after a letter has been added or removed from the word.
        if (wordChange)
        {
            al_draw_filled_rectangle(100 * scaleFactor, 85 * scaleFactor, (100 + maxLength * 10) * scaleFactor, 100 * scaleFactor, al_map_rgb(0, 0, 128));
            drawString(100 * scaleFactor, 85 * scaleFactor, word, textColour_);
            al_flip_display();
        }
    } while (chosenKey != ALLEGRO_KEY_ENTER && chosenKey != ALLEGRO_KEY_PAD_ENTER);
}