#include "Player.h"

// Global player object.
Player player;

void initPlayer(u8 x, u8 y)
{
    // Initialise actor variables.
    u8 roomX = x / SCREEN_TILES_X;
    u8 roomY = y / SCREEN_TILES_Y;

    // Set room and position information based on x,y values.
    player.object.position.x = (x % SCREEN_TILES_X) * TILE_WIDTH_BYTES;
    player.object.position.y = (y % SCREEN_TILES_Y) * TILE_HEIGHT;
    player.object.previousPosition.x = player.object.position.x;
    player.object.previousPosition.y = player.object.position.y;
    player.object.position.room = roomX + roomY * 3;
    player.object.previousPosition.room = player.object.position.room;
    player.object.width = PLAYER_WIDTH;
    player.object.height = PLAYER_HEIGHT;

    // No fish eaten initially.
    player.fishEaten = 0;

    // The player sprite defaults to right.
    player.direction = DIRECTION_RIGHT;

    // Set redraw to draw the actor by default.
    player.redraw = TRUE;
}

void drawPlayer()
{
    if (player.redraw)
    {
        // Draws the player using the actor struct.
        drawActor((Actor*)&player, allegroPlayerSprites[player.direction]);

        // Set redraw to false ready for the next change when it will be set to true.
        player.redraw = FALSE;
    }
}

void movePlayer()
{
    // Update the player velocity based on controls currently down.
    if (gameControls.controlFlags & CONTROL_LEFT_FLAG)
    {
        player.object.velocity.x = -1;
        player.direction = DIRECTION_LEFT;
    }

    if (gameControls.controlFlags & CONTROL_RIGHT_FLAG)
    {
        player.object.velocity.x = 1;
        player.direction = DIRECTION_RIGHT;
    }

    if (gameControls.controlFlags & CONTROL_UP_FLAG)
    {
        player.object.velocity.y = -2;
    }

    if (gameControls.controlFlags & CONTROL_DOWN_FLAG)
    {
        player.object.velocity.y = 2;
    }
}

void logicPlayer()
{
    // Save player current position before updating position based on velocity.
    // If you do this after checking velocity it results in the player only being partially drawn (or not at all) on screen.
    player.object.previousPosition.x = player.object.position.x;
    player.object.previousPosition.y = player.object.position.y;

    // Check the player is moving.
    if (player.object.velocity.y || player.object.velocity.x)
    {
        if (player.object.velocity.x > 0)
        {
            // Player going right.
            if (player.object.position.x + player.object.velocity.x <= SCREEN_X - player.object.width)
            {
                player.object.position.x += player.object.velocity.x;
            }
        }
        else if (player.object.velocity.x < 0)
        {
            // Player going left.
            if (player.object.position.x + player.object.velocity.x >= 0)
            {
                player.object.position.x += player.object.velocity.x;
            }
        }

        if (player.object.velocity.y > 0)
        {
            // Player going down.
            // Check whether player next position below is on screen.
            if (player.object.position.y + player.object.velocity.y <= SCREEN_Y_NO_HUD - player.object.height)
            {
                player.object.position.y += player.object.velocity.y;
            }
        }
        else if (player.object.velocity.y < 0)
        {
            // Going up.
            // Check whether player next position above is on screen.
            if (player.object.position.y + player.object.velocity.y >= 0)
            {
                player.object.position.y += player.object.velocity.y;
            }
        }

        // Reset velocity after movement.
        player.object.velocity.y = 0;
        player.object.velocity.x = 0;

        // Redraw the player at the new position if they have moved.
        if (player.object.position.x != player.object.previousPosition.x || player.object.position.y != player.object.previousPosition.y)
        {
            player.redraw = TRUE;
        }
    }

    // Check if the player has collided with an enemy for eating.
    playerEnemyCollision();
}

void playerEnemyCollision()
{
    u8 i = 0;
    Enemy* enemy = roomEnemies;

    for (i = 0; i < roomEnemyCount; ++i)
    {
        // Make sure the enemy has not previously been destroyed.
        if (enemy->health)
        {
            u8 xOverlap = difference(enemy->object.position.x, player.object.position.x);
            u8 yOverlap = difference(enemy->object.position.y, player.object.position.y);

            // Check that there is an overlap between player and enemy.
            if (xOverlap < enemy->object.width && yOverlap < enemy->object.height)
            {
                // Make sure there is a significant overlap between player and enemy before eating.
                if (xOverlap < 1)
                {
                    enemy->health = 0;

                    // Save the type of fish the player has eaten.
                    fishEaten[player.fishEaten] = enemy->type;

                    // Update the number of fish eaten once you have sufficient overlap.
                    ++player.fishEaten;

                    // Update player score on eating the fish.
                    levelScore += fishScore(enemy->type);

                    // Force redraw of the hud.
                    hudRedraw = TRUE;

                    if (player.fishEaten == MAX_FISH_CAN_EAT)
                    {
                        // Player has eaten the maximum amount and will now finish the level.
                        levelState = LEVEL_FULL;
                        gameRunning = FALSE;
                    }
                }
            }
        }

        // Move to the next enemy in the array.
        ++enemy;
    }
}