#include <allegro5/allegro.h>
#include <allegro5/allegro_font.h>
#include <allegro5/allegro_ttf.h>
#include <stdio.h>
#include "SpriteSets.h"
#include "Menu.h"
#include "Text.h"

// The ALLEGRO_DISPLAY representing the content of the screen.
ALLEGRO_DISPLAY* display_ = NULL;

// Locate the path to the configuration file.
// The file should be in the current directory or a parent directory.
ALLEGRO_PATH* locateConfigPath(const char* filename);

// Calculate the scale factor for the game based on the screen resolution.
unsigned char calculateScaleFactor(unsigned int screenWidth, unsigned int screenHeight);

int main()
{
    ALLEGRO_PATH* configPath = locateConfigPath("game.cfg");

    if (configPath == NULL)
    {
        return -1;
    }

    ALLEGRO_CONFIG* config = al_load_config_file(al_path_cstr(configPath, '/'));

    if (!al_init())
    {
        fprintf(stderr, "failed to initialise allegro!\n");

        return -2;
    }

    if (!al_init_font_addon())
    {
        fprintf(stderr, "failed to initialise font addon!\n");

        return -3;
    }

    if (!al_init_ttf_addon())
    {
        fprintf(stderr, "failed to initialise ttf addon!\n");

        return -4;
    }

    if (!al_init_primitives_addon())
    {
        fprintf(stderr, "failed to initialise primitives addon!\n");

        return -5;
    }

    int screenWidth = atoi(al_get_config_value(config, "", "display_width"));
    int screenHeight = atoi(al_get_config_value(config, "", "display_height"));

    // With default scaling the minimal screen size is 640x400.
    if (screenWidth < 640)
    {
        screenWidth = 640;
    }

    if (screenHeight < 400)
    {
        screenHeight = 400;
    }

    display_ = al_create_display(screenWidth, screenHeight);

    // Calculate the scale factor which will determine what multiple of 16 to use for sprites.
    scaleFactor = calculateScaleFactor(screenWidth, screenHeight);

    if (!display_)
    {
        fprintf(stderr, "failed to create display!\n");

        return -6;
    }

    initFonts(al_get_config_value(config, "", "font"));

    u8 fpsCap = atoi(al_get_config_value(config, "", "fps_cap"));

    // Initialise the sprite sets used by allegro.
    initAllAllegroSpriteSets();

    // Runs the game menu including menu loop.
    runMenu(fpsCap);

    // Cleanup allegro.
    destroyFonts();

    // Destroy the sprite sets used by allegro.
    destroyAllAllegroSpriteSets();

    al_shutdown_ttf_addon();
    al_shutdown_font_addon();
    al_shutdown_primitives_addon();
    al_destroy_display(display_);
    al_destroy_path(configPath);
    al_destroy_config(config);

    return 0;
}

ALLEGRO_PATH* locateConfigPath(const char* filename)
{
    // Start from the current directory and assume that the config file is in this directory.
    ALLEGRO_PATH* directory = al_create_path(al_get_current_directory());
    al_set_path_filename(directory, filename);

    // Make sure that the directory is still valid.
    while (al_get_path_num_components(directory) > 0)
    {
        // Check if the config file exists in the directory.
        if (al_filename_exists(al_path_cstr(directory, '/')))
        {
            return directory;
        }
        else
        {
            // Remove the part of the directory before the filename.
            al_remove_path_component(directory, -1);
        }
    }

    // Cleanup the directory created.
    al_destroy_path(directory);

    // Return NULL as the filename does not exist.
    return NULL;
}

unsigned char calculateScaleFactor(unsigned int screenWidth, unsigned int screenHeight)
{
    // The scale factor is the amount that the sprite width can be multiplied by (in multiples of 16).
    // 320x200 = 16->16
    // 640x400 = 16->32
    // 960x600 = 16->48
    // 1280x800 = 16->64
    // 1600x1000 = 16->80
    // 1920x1200 = 16->96
    unsigned char xScaleFactor = screenWidth / 320;
    unsigned char yScaleFactor = screenHeight / 200;

    if (xScaleFactor > yScaleFactor)
    {
        return yScaleFactor;
    }
    else
    {
        return xScaleFactor;
    }
}